package sprite;
import java.awt.Graphics;

/**
 * A Sprite represents a small graphical element that can draw itself in a SpritePanel.
 * This class is not ordinarily used directly for creating objects.  Use subclasses 
 * instead.  It does proved a large number of methods and properties for working with sprites.
 * Note that this class works closely with the SpritePanel class to implement sprite behavior.
 * A Sprite must be added to a SpritePanel for it to do anything useful.
 */
public class Sprite {
	
	/**
	 * A value for the exitPolicy property, indicating that the sprite should bounce off
	 * the edge of the SpritePanel when it hits it.  The sprite will not move outside the panel.
	 * This is the default behavior for a sprite.
	 */
	public static final int EXIT_POLICY_BOUNCE = 0;
	
	/**
	 * A value for the exitPolicy property, indicating that the sprite should stop moving if it
	 * hits the edge of the SpritPanel.  The sprite will not move outside the panel.
	 */
	public static final int EXIT_POLICY_STOP = 1;
	
	/**
	 * A value for the exitPolicy property, indicating that the sprite should die when it moves
	 * outside the SpritePanel.  The sprite does not die until it has moved completely outside
	 * the panel.
	 */
	public static final int EXIT_POLICY_DIE = 2;
	
	/**
	 * A value for the exitPolicy property, indicating that the sprite should wrap around to the
	 * opposite edge of the panel when it hits an edge of the SpritePanel.  After the sprite moves
	 * completely out of the panel, it reappears on the opposite side.
	 */
	public static final int EXIT_POLICY_WRAP = 3;

	
	private double x,y;   // The position of the center of the sprite, in pixels.
	private double dx,dy; // The velocity of the sprite, in pixels per frame.
	private int width,height;   // The size of the sprite, which is not guaranteed to match the actual picture.
	private SpritePanel home;  // The SpritePanel that contains this sprite, set automatically.
	private boolean dead = false;  // When set to true, the sprite is removed from its SpritePanel.
	private boolean immuneToBullets;   // if true, getting hit by a Bullet does not kill this sprite.
	private boolean useOvalForContains;  // if true, the shape used for collision detection is oval instead of rect.
	private int exitPolicy = EXIT_POLICY_BOUNCE;  // determines what happens to the sprite when it hits an edge.
	
	/**
	 *  This method is called by the SpritePanel to tell the sprite to draw itself.
	 *  It is not meant to be called directly.  Subclasses should override this method
	 *  to draw an appropriate picture.  The method in this class just draws a tiny
	 *  black dot at the location of the sprite.
	 */
	public void draw(Graphics g) {
		int xi = (int)x;
		int yi = (int)y;
		g.fillRect(xi-1,yi-1,2,2);
	}
	
	/**
	 * This method is called by the SpritePanel to tell that sprite that a new frame
	 * is about to be drawn.  It gives the sprite a chance to update its variables, if
	 * necessary.  Note that the movement of the sprite, as required by its velocity
	 * setting, is done elsewhere, not in this method.  This method is not meant to
	 * be called directly.  Subclasses can override this method if they need to do 
	 * something for each frame; however, that is probably relatively rare.  The
	 * method in this clas does nothing.
	 */
	public void prepareForNextFrame(int frameNumber) {
	}
	
	/**
	 * This method is called by the SpritePanel when the sprite dies for any reason.
	 * It is not meant to be called directly.  The method in this class does nothing.
	 * Subclasses that need to perform some action when the sprite dies can redefine
	 * this method.
	 */
	public void onDeath() {
	}
	
	public double getX() {
		return x;
	}

	public double getY() {
		return y;
	}

	public double getDx() {
		return dx;
	}

	public double getDy() {
		return dy;
	}
	
	/**
	 * Sets the x-coordinate of the center of the sprite.
	 */
	public void setX(double x) {
		this.x = x;
	}

	/**
	 * Sets the y-coordinate of the center of the sprite.
	 */
	public void setY(double y) {
		this.y = y;
	}

	/**
	 * Sets the horizontal (x) component of the sprite's velocity,
	 * given in pixels per frame.  The velocity is zero by default.
	 */
	public void setDx(double dx) {
		this.dx = dx;
	}

	/**
	 * Sets the vertical (y) component of the sprite's velocity,
	 * given in pixels per frame.  The velocity is zero by default.
	 */
	public void setDy(double dy) {
		this.dy = dy;
	}

	/**
	 * Sets the location of the center of the sprite.  Note that there
	 * is nothing to stop you from setting the location to be outside
	 * the panel,
	 */
	public void setLocation(double newX, double newY) {
		x = newX;
		y = newY;
	}
	
	/**
	 * Sets the velocity of the sprite.  The default velocity is zero.
	 * If the sprite is going to move by itself, a non-zero velocity
	 * must be assigned to it.
	 * @param newDx the horizontal component of the velocity
	 * @param newDy the vertical component of the velocity
	 */
	public void setVelocity(double newDx, double newDy) {
		dx = newDx;
		dy = newDy;
	}
	
	/**
	 * Sets the velocity of the sprite to a random value.  The direction
	 * in which the sprite should move is picked at random.  The speed
	 * is also set at random, based on the parameters.
	 * @param minimumSpeed the minimum possible speed.  Must be greater than or equal to zero.
	 * @param maximumSpeed the maximum possible speed.  Must be greater than or equal to
	 * the minimum speed.  (If they are equal, then the speed is not random, but the direction
	 * is still random.)
	 */
	public void randomVelocity(double minimumSpeed, double maximumSpeed) {
		if (minimumSpeed < 0)
			throw new IllegalArgumentException("Speed can't be negative");
		if (maximumSpeed < minimumSpeed)
			throw new IllegalArgumentException("Maximum speed can't be less than minimum speed");
		double speed = minimumSpeed + (maximumSpeed - minimumSpeed)*Math.random();
		double angle = 2*Math.PI*Math.random();
		dx = speed * Math.cos(angle);
		dy = speed * Math.sin(angle);
	}
	
	public int getSpriteWidth() {
		return width;
	}

	public int getSpriteHeight() {
		return height;
	}

	/**
	 * Sets the width of the sprite.  This should be equal to the width of the picture that
	 * the sprite draws, but that is not enforced.  The width and height are used in two
	 * cases:  For testing whether the sprite contains a given point (which is used to tell
	 * whether the sprite has been hit by a bullet) and for doing the
	 * "exitPolicy" calculations (that is, testing whether the sprite has hit or passed
	 * the boundary of the SpritePanel).  Of course, the width and height might be used
	 * for other purposes in subclasses.  But if the width isn't used for anything, then
	 * there is no need to set it!  The default value is zero.
	 * @param width the new width of the sprite.  Must be greater than or equal to zero.
	 */
	public void setSpriteWidth(int width) {
		if (width < 0)
			throw new IllegalArgumentException("Width of a sprite can't be negative.");
		this.width = width;
	}

	/**
	 * Sets the height of the sprite.  This should be equal to the height of the picture that
	 * the sprite draws, but that is not enforced.  The width and height are used in two
	 * cases:  For testing whether the sprite contains a given point (which is used to tell
	 * whether the sprite has been hit by a bullet) and for doing the
	 * "exitPolicy" calculations (that is, testing whether the sprite has hit or passed
	 * the boundary of the SpritePanel).  Of course, the width and height might be used
	 * for other purposes in subclasses.  But if the height isn't used for anything, then
	 * there is no need to set it!  The default value is zero.
	 * @param width the new width of the sprite.  Must be greater than or equal to zero.
	 */
	public void setSpriteHeight(int height) {
		if (width < 0)
			throw new IllegalArgumentException("Height of a sprite can't be negative.");
		this.height = height;
	}
	
	public int getExitPolicy() {
		return exitPolicy;
	}

	/**
	 * Tells what happens when a moving sprite reaches the edge of the SpritePanel
	 * that contains it.  This applies only to sprites that have non-zero velocity.
	 * The exitPolicy does not stop you from placing a living, non-moving sprite outside
	 * the boundaries of the SpritePanel.
	 * @param policyCode the code for the policy. It must be one of the constants:
	 * Sprite.EXIT_POLICY_BOUNCE (the sprite bounces off the edge of the SpritePanel),
	 * Sprite.EXIT_POLICY_STOP (the sprite stops moving when it hits the edge of the SpritePanel),
	 * Sprite.EXIT_POLICY_DIE (the sprite dies when it leaves the SpritePanel), or
	 * Sprite.EXIT_POLICY_WRAP (the sprite re-emerges at the opposite edge when it leaves the SpritePanel),
	 * 
	 */
	public void setExitPolicy( int policyCode ) {
		if (policyCode < 0 || policyCode > 4)
			throw new IllegalArgumentException("Illegal exit policy code");
		exitPolicy = policyCode;
	}

	/**
	 * Calling this method kills the sprite.  It is called automatically in some cases, such as
	 * when a moving sprite with exitPolicy = Sprite.EXIT_POLICY_EXIT moves outside the SpritePanel.
	 * When a sprite dies, it is automatically removed from the SpritePanel that contained it,
	 * and it is no longer drawn on the screen.  A dead sprite cannot be brought back to life.
	 */
	public void die() {
		dead = true;
		onDeath();
	}
	
	/**
	 * Tells whether the sprite is dead.
	 */
	public boolean isDead() {
		return dead;
	}
	
	/**
	 * Tests whether the sprite contains the point (a,b).  The test uses the
	 * width and height of the sprite as set by setSpriteWidth() and setSpriteHeight().
	 * These values are not guaranteed to represent the actual size of the picture drawn
	 * by the sprite; it is up to subclasses to make sure that is true.  This method is
	 * used by the SpritePanel to test whether a bullet has hit the sprite.  Note that the
	 * sprite is assumed to be a rectangle, unless you call sprite.setUseOvalForContains(true).
	 */
	public boolean contains(double a, double b) {
		if (!useOvalForContains || width == 0 || height == 0) {
			return a >= x-width/2 && a <= x+width/2 && b > y-height/2 && b < y+height/2;
		}
		else {
		   return Math.pow( (a-x)/(width/2.0), 2) + Math.pow( (b-y)/(height/2.0), 2) <= 1;
		}
	}

	/**
	 * Tells whether the shape of this sprite is oval or rectangular.  The value is used
	 * only in the method contains(a,b).
	 * @param useOvalForContains true says that the shape is oval.
	 */
	public void setUseOvalForContains(boolean useOvalForContains) {
		this.useOvalForContains = useOvalForContains;
	}

	/**
	 *  Says whether this sprite should be immune to bullets.  A sprite that is not immune to
	 *  bullets dies when it is hit by a bullet.
	 */
	public void setImmuneToBullets(boolean immune) {
		immuneToBullets = immune;
	}
	
	public boolean isImmuneToBullets() {
		return immuneToBullets;
	}
	
	/**
	 * Gets the SpritePanel that contains this sprite.  That panel is called the "home" of the
	 * sprite.  The value of home is set automatically when the sprite is added to a SpritePanel.
	 * @return the panel, or null if this sprite is not in a SpritePanel
	 */
	public SpritePanel getHome() {
		return home;
	}

	void setHome(SpritePanel panel) { // Package-private method used by SpritePanel.
		home = panel;
	}

	void moveToNextFramePosition() { // Package-private method updates the sprite's position in each frame.
		if (dx == 0 && dy == 0)
			return;
		x += dx;
		y += dy;
		double left =  x - width/2;
		double top =  y - height/2;
		double right =  x + width/2;
		double bottom =  y + height/2;
		switch (exitPolicy) {
		case EXIT_POLICY_BOUNCE:
			if (right > home.getWidth()) {
				dx = -Math.abs(dx);
				x = x - 2*(right - home.getWidth());
			}
			if (bottom > home.getHeight()) {
				dy = -Math.abs(dy);
				y = y - 2*(bottom - home.getHeight());
			}
			if (left < 0) {
				dx = Math.abs(dx);
				x = x - 2*left;
			}
			if (top < 0) {
				dy = Math.abs(dy);
				y = y - 2*top;
			}
			break;
		case EXIT_POLICY_STOP:
			if (right > home.getWidth()) {
				x = home.getWidth() - width/2;
				y -= dy;
			}
			if (bottom > home.getHeight()) {
				y = home.getHeight() - height/2;
				x -= dx;
			}
			if (left < 0) {
				x = width/2;
				y -= dy;
			}
			if (top < 0) {
				y = height/2;
				x -= dx;
			}
			break;
		case EXIT_POLICY_DIE:
			if (right <= 0 || bottom <= top || left > home.getWidth() || top > home.getHeight())
				die();
			break;
		case EXIT_POLICY_WRAP:
			if (right < 0)
				x = home.getWidth() + width/2;
			if (left > home.getWidth())
				x = -width/2;
			if (bottom < 0)
				y = home.getHeight() + height/2;
			if (top > home.getHeight())
				y = -height/2;
			break;
		}
	}
	
}
