package sprite;

import java.awt.*;

/**
 * A sprite that displays a string of text.  The position of 
 * the sprite is the center of the string.
 */
public class TextSprite extends Sprite{
	
	private String text;     // The text to be drawn, or null if there is no text.
	private Color color;     // the color of the text
	private Font font;       // The font to be used; if null, the SpritePane's default font is used.
	private int textAscent;  //
	
	/**
	 * Create a text sprite to display given text.  Note that the width
	 * and height of the sprite will be computed automatically, based on
	 * the font and the text.
	 * @throws IllegalArgumentException if s is null
	 */
	public TextSprite(String s) {
		setText(s);
		color = Color.BLACK;
	}
	
	/**
	 * Draw the text, centered on the (x,y)-position of the sprite.
	 */
	public void draw(Graphics g) {
		if (getSpriteWidth() == 0) {
			   // We can't determine the width and height until we have
			   // a graphics context.
			if (font == null)
				font = g.getFont();
			FontMetrics fm = g.getFontMetrics(font);
			setSpriteWidth( fm.stringWidth(text) );
			setSpriteHeight( fm.getAscent() + fm.getDescent() );
			textAscent = fm.getAscent();
		}
		int x = (int)getX();
		int y = (int)getY();
		g.setColor(color);
		g.setFont(font);
		g.drawString(text, x-getSpriteWidth()/2, y-getSpriteHeight()/2 + textAscent);
	}

	public String getText() {
		return text;
	}

	public Color getColor() {
		return color;
	}

	public Font getFont() {
		return font;
	}

	/**
	 * Changes the text that is displayed by this Sprite.
	 * @throws IllegalArgumentException if s is null.
	 */
	public void setText(String s) {
		if (s == null || s.length() == 0)
			throw new IllegalArgumentException("Can't have empty text.");
		text = s;
		setSpriteWidth(0);  // will be recomputed in next call to draw();
	}

	/**
	 * Sets the color of this Sprite.  If the specified color is null,
	 * the default color, black, is used.
	 */
	public void setColor(Color c) {
		if (c == null )
			color = Color.BLACK;
		else
			color = c;
	}

	/**
	 * Changes the font to be used for drawing the text.
	 */
	public void setFont(Font f) {
		font = f;
		setSpriteWidth(0);  // will be recomputed in next call to draw();
	}

}
